# Generated by Django 3.1.7 on 2021-03-22 04:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='reportcolumn',
            options={'verbose_name': 'Column', 'verbose_name_plural': 'Columns'},
        ),
        migrations.AlterOrderWithRespectTo(
            name='reportcolumn',
            order_with_respect_to='report',
        ),
    ]



import htmlgenerator as hg
from django import forms
from django.http import HttpResponseNotAllowed

from .. import layout as _layout  # prevent name clashing
from ..forms.fields import FormsetField
from ..forms.forms import breadmodelform_factory
from .edit import EditView


# Read view is the same as the edit view but form elements are disabled
class ReadView(EditView):
    """TODO: documentation"""

    accept_global_perms = True
    fields = None
    urlparams = (("pk", int),)

    def post(self, *args, **kwargs):
        return HttpResponseNotAllowed(["GET"])

    def get_form_class(self, form=forms.models.ModelForm):
        return breadmodelform_factory(
            request=self.request,
            model=self.model,
            layout=self._get_layout_cached(),
            instance=self.object,
            baseformclass=form,
            cache_querysets=False,
        )

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        for field in form.fields.values():
            field.disabled = True
            if isinstance(field, FormsetField):
                for subfield in field.formsetclass.form.base_fields.values():
                    subfield.disabled = True
        return form

    def get_layout(self):
        return layoutasreadonly(super().get_layout())

    def get_required_permissions(self, request):
        return [f"{self.model._meta.app_label}.view_{self.model.__name__.lower()}"]


def layoutasreadonly(layout):
    layout.wrap(
        lambda element, ancestors: isinstance(element, _layout.form.Form)
        and element.standalone,
        hg.FIELDSET(readonly="true"),
    )

    layout.delete(
        lambda element, ancestors: any(
            [isinstance(a, _layout.form.Form) for a in ancestors]
        )
        and (
            (
                isinstance(
                    element,
                    hg.BUTTON,
                )
                and "bx--tag" not in element.attributes.get("_class", "")
            )
            or getattr(element, "attributes", {}).get("type") == "submit"
        )
    )
    for form in layout.filter(lambda element, ancestors: isinstance(element, hg.FORM)):
        form.attributes["onsubmit"] = "return false;"
    return layout



import htmlgenerator


class Grid(htmlgenerator.DIV):
    MODES = ("narrow", "condensed", "full-width")

    def __init__(self, *children, gridmode="full-width", gutter=True, **attributes):
        attributes["_class"] = attributes.get("_class", "") + " bx--grid"
        if gridmode is not None:
            if gridmode not in Grid.MODES:
                raise ValueError(f"argument 'gridmode' must be one of {Grid.MODES}")
            attributes["_class"] += f" bx--grid--{gridmode}"
        if not gutter:
            attributes["_class"] += " bx--no-gutter"
        super().__init__(*children, **attributes)


class Row(htmlgenerator.DIV):
    def __init__(self, *children, gridmode=None, **attributes):
        attributes["_class"] = attributes.get("_class", "") + " bx--row"
        if gridmode is not None:
            attributes["_class"] += f" bx--row--{gridmode}"
        super().__init__(*children, **attributes)


class Col(htmlgenerator.DIV):
    def __init__(self, *children, breakpoint="lg", width=None, **attributes):
        """
        breakpoint: Can be one of "sm", "md", "lg", "xlg", "max"
        """
        colclass = "bx--col"
        if width is not None:
            colclass += f"-{breakpoint}-{width}"
        attributes["_class"] = attributes.get("_class", "") + f" {colclass}"
        super().__init__(*children, **attributes)



import htmlgenerator as hg
from django import forms
from django.utils.html import mark_safe
from django.utils.translation import gettext_lazy as _

from .button import Button
from .datatable import DataTable, DataTableColumn
from .email_input import EmailInput
from .notification import InlineNotification
from .phone_number_input import PhoneNumberInput
from .url_input import UrlInput


class Form(hg.FORM):
    @staticmethod
    def from_django_form(form, **kwargs):
        return Form.from_fieldnames(form, form.fields, **kwargs)

    @staticmethod
    def from_fieldnames(form, fieldnames, **kwargs):
        return Form.wrap_with_form(
            form, *[FormField(fieldname) for fieldname in fieldnames], **kwargs
        )

    @staticmethod
    def wrap_with_form(form, *elements, submit_label=None, **kwargs):
        if kwargs.get("standalone", True) is True:
            elements += (
                hg.DIV(
                    Button(submit_label or _("Save"), type="submit"),
                    _class="bx--form-item",
                    style="margin-top: 2rem",
                ),
            )
        return Form(form, *elements, **kwargs)

    def __init__(self, form, *children, use_csrf=True, standalone=True, **attributes):
        """
        form: lazy evaluated value which should resolve to the form object
        children: any child elements, can be formfields or other
        use_csrf: add a CSRF input, but only for POST submission and standalone forms
        standalone: if true, will add a CSRF token and will render enclosing FORM-element
        """
        self.form = form
        self.standalone = standalone
        defaults = {"method": "POST", "autocomplete": "off"}
        defaults.update(attributes)
        if (
            defaults["method"].upper() == "POST"
            and use_csrf is not False
            and standalone is True
        ):
            children = (CsrfToken(),) + children

        super().__init__(*children, **defaults)

    def formfieldelements(self):
        return self.filter(
            lambda elem, parents: isinstance(elem, FormChild)
            and not any((isinstance(p, Form) for p in parents[1:]))
        )

    def render(self, context):
        form = hg.resolve_lazy(self.form, context)
        for formfield in self.formfieldelements():
            formfield.form = form
        for error in form.non_field_errors():
            self.insert(0, InlineNotification(_("Form error"), error, kind="error"))
        for hidden in form.hidden_fields():
            for error in hidden.errors:
                self.insert(
                    0,
                    InlineNotification(
                        _("Form error: "), f"{hidden.name}: {error}", kind="error"
                    ),
                )
        if self.standalone:
            if form.is_multipart() and "enctype" not in self.attributes:
                self.attributes["enctype"] = "multipart/form-data"
            return super().render(context)
        return super().render_children(context)


class FormChild:
    """Used to mark elements which need the "form" attribute set by the parent form before rendering"""


class FormField(FormChild, hg.BaseElement):
    """Dynamic element which will resolve the field with the given name
    and return the correct HTML, based on the widget of the form field or on the passed argument 'fieldtype'"""

    def __init__(
        self,
        fieldname,
        fieldtype=None,
        hidelabel=False,
        elementattributes=None,
        widgetattributes=None,
        formname="form",
    ):
        if fieldtype is not None and not isinstance(fieldtype, type):
            raise ValueError("argument 'fieldtype' is not a type")
        self.fieldname = fieldname
        self.fieldtype = fieldtype
        self.widgetattributes = widgetattributes or {}
        self.elementattributes = elementattributes or {}
        self.hidelabel = hidelabel
        self.form = None  # will be set by the render method of the parent method
        self.formname = formname  # in the future we should only depend on the formname to extract the form from the context

    def render(self, context):
        form = self.form or hg.resolve_lazy(context[self.formname], context)
        element = _mapwidget(
            form[self.fieldname],
            self.fieldtype,
            self.elementattributes,
            self.widgetattributes,
        )
        if self.hidelabel:
            element.replace(
                lambda e, ancestors: isinstance(e, hg.LABEL), None, all=True
            )
        return element.render(context)


class FormsetField(hg.Iterator):
    def __init__(
        self,
        fieldname,
        content,
        formname="form",
        formsetinitial=None,
        **formsetfactory_kwargs,
    ):
        self.fieldname = fieldname
        self.formname = formname
        self.formsetfactory_kwargs = formsetfactory_kwargs  # used in bread.forms.forms._generate_formset_class, maybe refactor this?
        self.formsetinitial = formsetinitial  # used in bread.forms.forms._generate_formset_class, maybe refactor this?
        self.content = content
        if isinstance(self.content, FormField):
            self.content = hg.BaseElement(self.content)

        # search fields which have explicitly been defined in the content element
        declared_fields = set(
            f.fieldname
            for f in self.content.filter(lambda e, ancestors: isinstance(e, FormField))
        )

        # append all additional fields of the form which are not rendered explicitly
        # These should be internal, hidden fields (can we test this somehow?)
        self.content.append(
            hg.F(
                lambda c: hg.BaseElement(
                    *[
                        FormField(field, formname="formset_form")
                        for field in c[self.formname][
                            self.fieldname
                        ].formset.empty_form.fields
                        if field not in declared_fields
                        and field != forms.formsets.DELETION_FIELD_NAME
                    ]
                )
            )
        )

        super().__init__(
            iterator=hg.C(f"{self.formname}.{self.fieldname}.formset"),
            loopvariable="formset_form",
            content=Form(hg.C("formset_form"), self.content, standalone=False),
        )

    @property
    def management_form(self):
        # the management form is required for Django formsets
        return hg.BaseElement(
            # management forms, for housekeeping of inline forms
            hg.F(
                lambda c: Form.from_django_form(
                    c[self.formname][self.fieldname].formset.management_form,
                    standalone=False,
                )
            ),
            # Empty form as template for new entries. The script tag works very well
            # for this since we need a single, raw, unescaped HTML string
            hg.SCRIPT(
                Form(
                    hg.C(f"{self.formname}.{self.fieldname}.formset.empty_form"),
                    hg.WithContext(
                        self.content,
                        formset_form=hg.C(
                            f"{self.formname}.{self.fieldname}.formset.empty_form"
                        ),
                    ),
                    standalone=False,
                ),
                id=hg.BaseElement(
                    "empty_",
                    hg.C(f"{self.formname}.{self.fieldname}.formset.prefix"),
                    "_form",
                ),
                type="text/plain",
            ),
            hg.SCRIPT(
                mark_safe(
                    "document.addEventListener('DOMContentLoaded', e => init_formset('"
                ),
                hg.C(f"{self.formname}.{self.fieldname}.formset.prefix"),
                mark_safe("'));"),
            ),
        )

    def add_button(self, container_css_selector, label=_("Add"), **kwargs):
        prefix = hg.C(f"{self.formname}.{self.fieldname}.formset.prefix")
        defaults = {
            "icon": "add",
            "notext": True,
            "buttontype": "tertiary",
            "id": hg.BaseElement(
                "add_", prefix, "_button"
            ),  # required for javascript to work correctly
            "onclick": hg.BaseElement(
                "formset_add('",
                prefix,
                "', '",
                container_css_selector,
                "');",
            ),
        }
        return Button(label, **{**defaults, **kwargs})

    @staticmethod
    def as_plain(*args, add_label=_("Add"), **kwargs):
        """Shortcut to render a complete formset with add-button"""
        formset = FormsetField(*args, **kwargs)
        id = hg.html_id(formset, prefix="formset-")
        return hg.BaseElement(
            hg.DIV(formset, id=id),
            formset.management_form,
            formset.add_button(
                buttontype="ghost",
                notext=False,
                label=add_label,
                container_css_selector=f"#{id}",
            ),
        )

    @staticmethod
    def as_datatable(
        fieldname,
        fields,
        title=None,
        formname="form",
        formsetfield_kwargs=None,
        **kwargs,
    ):
        """
        :param str fieldname: The fieldname which should be used for an formset, in general a one-to-many or many-to-many field
        :param list fields: A list of strings or objects. Strings are converted to DataTableColumn, objects are passed on as they are
        :param str title: Datatable title, automatically generated from form if None
        :param str formname: Name of the surounding django-form object in the context
        :param dict formsetfield_kwargs: Arguments to be passed to the FormSetField constructor
        :param kwargs: Arguments to be passed to the DataTable constructor
        :return: A datatable with inline-editing capabilities
        :rtype: hg.HTMLElement
        """

        columns = []
        for f in fields:
            if isinstance(f, str):
                f = DataTableColumn(
                    hg.C(f"{formname}.{fieldname}.formset.form.base_fields.{f}.label"),
                    FormField(f, hidelabel=True),
                )
            columns.append(f)
        columns.append(
            DataTableColumn(
                _("Order"),
                hg.If(
                    hg.C(f"{formname}.{fieldname}.formset.can_order"),
                    FormField(forms.formsets.ORDERING_FIELD_NAME, hidelabel=True),
                ),
            )
        )
        columns.append(
            DataTableColumn(
                "",
                hg.If(
                    hg.C(f"{formname}.{fieldname}.formset.can_delete"),
                    InlineDeleteButton(parentcontainerselector="tr"),
                ),
            )
        )

        formset = FormsetField(
            fieldname,
            DataTable.row(columns),
            formname=formname,
            **(formsetfield_kwargs or {}),
        )
        id = hg.html_id(formset, prefix="formset-")

        return hg.BaseElement(
            DataTable(
                row_iterator=formset,
                rowvariable="",
                columns=columns,
                id=id,
                **kwargs,
            ).with_toolbar(
                title=title or hg.C(f"{formname}.{fieldname}.label"),
                primary_button=formset.add_button(
                    buttontype="primary", container_css_selector=f"#{id} tbody"
                ),
            ),
            formset.management_form,
        )


class InlineDeleteButton(FormChild, Button):
    def __init__(self, parentcontainerselector, label=_("Delete"), **kwargs):
        """
        Show a delete button for the current inline form. This element needs to be inside a FormsetField
        parentcontainerselector: CSS-selector which will be passed to element.closest in order to select the parent container which should be hidden on delete
        """
        defaults = {
            "notext": True,
            "small": True,
            "icon": "trash-can",
            "buttontype": "ghost",
            "onclick": f"delete_inline_element(this.querySelector('input[type=checkbox]'), this.closest('{parentcontainerselector}'))",
        }
        defaults.update(kwargs)
        super().__init__(
            label,
            FormField(
                forms.formsets.DELETION_FIELD_NAME,
                elementattributes={"style": "display: none"},
            ),
            **defaults,
        )


class HiddenInput(FormChild, hg.INPUT):
    def __init__(self, fieldname, widgetattributes, boundfield=None, **attributes):
        self.fieldname = fieldname
        super().__init__(type="hidden", **{**widgetattributes, **attributes})

        self.attributes["id"] = boundfield.auto_id
        if boundfield is not None:
            self.attributes["name"] = boundfield.html_name
            if boundfield.value() is not None:
                self.attributes["value"] = boundfield.value()


class CsrfToken(FormChild, hg.INPUT):
    def __init__(self):
        super().__init__(type="hidden")

    def render(self, context):
        self.attributes["name"] = "csrfmiddlewaretoken"
        self.attributes["value"] = context["csrf_token"]
        return super().render(context)


def _mapwidget(
    field, fieldtype, elementattributes=None, widgetattributes=None, only_initial=False
):
    from .checkbox import Checkbox
    from .date_picker import DatePicker
    from .file_uploader import FileUploader
    from .multiselect import MultiSelect
    from .select import Select
    from .text_area import TextArea
    from .text_input import PasswordInput, TextInput

    widgetattributes = update_widgetattributes(field, only_initial, widgetattributes)
    elementattributes = {
        "label": field.label,
        "help_text": field.help_text,
        "errors": field.errors,
        "disabled": field.field.disabled,
        "required": field.field.required,
        **getattr(field.field, "layout_kwargs", {}),
        **(elementattributes or {}),
    }

    if fieldtype:
        return fieldtype(
            fieldname=field.name,
            widgetattributes=widgetattributes,
            boundfield=field,
            **elementattributes,
        )

    if isinstance(field.field.widget, forms.CheckboxInput):
        widgetattributes["checked"] = field.value()
        return hg.DIV(
            Checkbox(
                widgetattributes=widgetattributes,
                **elementattributes,
            ),
            _class="bx--form-item",
        )

    if isinstance(field.field.widget, forms.CheckboxSelectMultiple):
        del elementattributes["label"]
        del elementattributes["required"]

        return hg.DIV(
            *[
                Checkbox(
                    **elementattributes,
                    label=widget.data["label"],
                    widgetattributes={
                        **widgetattributes,
                        "name": widget.data["name"],
                        "value": widget.data["value"],
                        "checked": widget.data["selected"],
                        **widget.data["attrs"],
                    },
                )
                for widget in field.subwidgets
            ],
            _class="bx--form-item",
        )

    if isinstance(field.field.widget, forms.Select):
        # This is to prevent rendering extrem long lists of database entries
        # (e.g. all persons) for relational fields if the field is disabled.
        if isinstance(field.field.widget, forms.SelectMultiple):
            if field.field.disabled and hasattr(field.field, "queryset"):
                field.field.queryset = field.field.queryset.filter(
                    pk__in=[i.pk for i in field.form.initial.get(field.name)]
                )
            return hg.DIV(
                MultiSelect(
                    field.field.widget.optgroups(
                        field.name,
                        field.field.widget.get_context(field.name, field.value(), {})[
                            "widget"
                        ]["value"],
                    ),
                    widgetattributes=widgetattributes,
                    **elementattributes,
                ),
                _class="bx--form-item",
            )
        if field.field.disabled and hasattr(field.field, "queryset"):
            default = field.form.initial.get(field.name) or None
            field.field.queryset = field.field.queryset.filter(
                pk=getattr(default, "pk", default)
            )
        return hg.DIV(
            Select(
                field.field.widget.optgroups(
                    field.name,
                    field.field.widget.get_context(field.name, field.value(), {})[
                        "widget"
                    ]["value"],
                ),
                widgetattributes=widgetattributes,
                **elementattributes,
            ),
            _class="bx--form-item",
        )

    WIDGET_MAPPING = {
        forms.TextInput: TextInput,
        # Attention: NumberInput is not the widget that is used for phone numbers. See below for handling of phone numbers
        forms.NumberInput: TextInput,
        forms.EmailInput: EmailInput,
        forms.URLInput: UrlInput,
        forms.PasswordInput: PasswordInput,
        forms.HiddenInput: HiddenInput,
        forms.DateInput: DatePicker,
        forms.DateTimeInput: TextInput,  # TODO
        forms.TimeInput: TextInput,  # TODO HIGH
        forms.Textarea: TextArea,
        forms.CheckboxInput: Checkbox,
        forms.NullBooleanSelect: Select,
        forms.SelectMultiple: MultiSelect,  # TODO HIGH
        forms.RadioSelect: TextInput,  # TODO HIGH
        forms.FileInput: FileUploader,
        forms.ClearableFileInput: FileUploader,  # TODO HIGH
        forms.MultipleHiddenInput: TextInput,  # TODO
        forms.SplitDateTimeWidget: TextInput,  # TODO
        forms.SplitHiddenDateTimeWidget: TextInput,  # TODO
        forms.SelectDateWidget: TextInput,  # TODO
    }

    fieldtype = (
        fieldtype
        or getattr(field.field, "layout", None)
        # needs to be above the WIDGET_MAPPING, because the field.field.widget is a forms.TextInput which would match
        or (
            PhoneNumberInput
            if hasattr(field.field.widget, "input_type")
            and field.field.widget.input_type == "tel"
            else None
        )
        or WIDGET_MAPPING[type(field.field.widget)]
    )
    if isinstance(fieldtype, type) and issubclass(fieldtype, hg.BaseElement):
        ret = fieldtype(
            fieldname=field.name,
            widgetattributes=widgetattributes,
            boundfield=field,
            **elementattributes,
        )
    else:
        ret = hg.DIV(f"Field {field.name}")

    if (
        field.field.show_hidden_initial and fieldtype != HiddenInput
    ):  # special case, prevent infinte recursion
        return hg.BaseElement(
            ret,
            _mapwidget(field, HiddenInput, only_initial=True),
        )

    return ret


def update_widgetattributes(field, only_initial, widgetattributes):
    # TODO: This can be simplified, and improved
    widgetattributes = widgetattributes or {}
    attrs = dict(field.field.widget.attrs)
    attrs.update(widgetattributes)
    attrs = field.build_widget_attrs(attrs)
    if getattr(field.field.widget, "allow_multiple_selected", False):
        attrs["multiple"] = True
    if field.auto_id and "id" not in field.field.widget.attrs:
        attrs.setdefault("id", field.html_initial_id if only_initial else field.auto_id)
    if "name" not in attrs:
        attrs["name"] = field.html_initial_name if only_initial else field.html_name
    value = field.field.widget.format_value(field.value())
    if value is not None and "value" not in attrs:
        attrs["value"] = value
    return attrs



import htmlgenerator as hg

from .helpers import ErrorListElement, HelpTextElement, LabelElement
from .icon import Icon


class Select(hg.DIV):
    def __init__(
        self,
        optgroups,
        light=False,
        inline=False,
        widgetattributes=None,
        label=None,
        help_text=None,
        errors=None,
        disabled=None,
        required=None,
        **attributes,
    ):
        """
        optgroups: list of 2-tuples: (group_name, group_items)
                   group_items: list of objects with attributes ['label', 'value', 'attrs'], attrs is a dict with html attributes

        """

        _class = attributes.pop("_class", "")
        widgetattributes = widgetattributes or {}
        widgetattributes["_class"] = (
            widgetattributes.get("_class", "") + " bx--select-input"
        )
        widgetattributes.setdefault("id", hg.html_id(self))

        select_wrapper = hg.DIV(
            hg.SELECT(
                hg.Iterator(
                    optgroups,
                    "optgroup",
                    hg.If(
                        hg.C("optgroup.0"),
                        hg.OPTGROUP(
                            hg.Iterator(
                                hg.C("optgroup.1"),
                                "option",
                                hg.OPTION(
                                    hg.C("option.label"),
                                    _class="bx--select-option",
                                    value=hg.C("option.value"),
                                    lazy_attributes=hg.C("option.attrs"),
                                ),
                            ),
                            _class="bx--select-optgroup",
                            label=hg.C("optgroup.0"),
                        ),
                        hg.Iterator(
                            hg.C("optgroup.1"),
                            "option",
                            hg.OPTION(
                                hg.C("option.label"),
                                _class="bx--select-option",
                                value=hg.C("option.value"),
                                lazy_attributes=hg.C("option.attrs"),
                            ),
                        ),
                    ),
                ),
                **widgetattributes,
            ),
            Icon(
                "chevron--down", size=16, _class="bx--select__arrow", aria_hidden="true"
            ),
            hg.If(
                errors,
                Icon(
                    "warning--filled",
                    size=16,
                    _class="bx--select__invalid-icon",
                ),
            ),
            _class="bx--select-input__wrapper",
            data_invalid=hg.If(errors, True),
        )

        super().__init__(
            LabelElement(
                label,
                _for=widgetattributes["id"],
                required=required,
                disabled=disabled,
            ),
            hg.If(
                inline,
                hg.DIV(
                    select_wrapper,
                    ErrorListElement(errors),
                    _class="bx--select-input--inline__wrapper",
                ),
                select_wrapper,
            ),
            HelpTextElement(help_text),
            hg.If(
                inline, None, ErrorListElement(errors)
            ),  # not displayed if this is inline
            _class=hg.BaseElement(
                _class,
                " bx--select",
                hg.If(inline, " bx--select--inline"),
                hg.If(light, " bx--select--light"),
                hg.If(errors, " bx--select--invalid"),
                hg.If(disabled, " bx--select--disabled"),
            ),
            **attributes,
        )



#!/usr/bin/env python

import sys

import django
from django.conf import settings

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "guardian",
    "bread",
    "bread.contrib.reports",
    "bread.contrib.workflows",
]

settings.configure(  # nosec because this is only for local development
    DEBUG=True,
    USE_TZ=True,
    USE_I18N=True,
    DATABASES={"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}},
    MIDDLEWARE_CLASSES=(),
    SITE_ID=1,
    INSTALLED_APPS=INSTALLED_APPS,
    AUTHENTICATION_BACKENDS=(
        "django.contrib.auth.backends.ModelBackend",
        "guardian.backends.ObjectPermissionBackend",
    ),
    SECRET_KEY="SECRET_KEY_FOR_TESTING",
    STATIC_URL="static/",
)

django.setup()

if __name__ == "__main__":
    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



from typing import Iterator, NamedTuple

import htmlgenerator as hg
from django.conf import settings
from django.core.paginator import Paginator
from django.utils.translation import gettext_lazy as _

from bread.utils.urls import link_with_urlparameters

from ..base import aslink_attributes
from .icon import Icon
from .select import Select


class PaginationConfig(NamedTuple):
    paginator: Paginator
    items_per_page_options: Iterator = getattr(
        settings, "DEFAULT_PAGINATION_CHOICES", [25, 50, 100]
    )
    page_urlparameter: str = (
        "page"  # URL parameter which holds value for current page selection
    )
    itemsperpage_urlparameter: str = (
        "itemsperpage"  # URL parameter which selects items per page
    )


class Pagination(hg.DIV):
    def __init__(
        self,
        paginator,
        items_per_page_options,
        page_urlparameter="page",
        itemsperpage_urlparameter="itemsperpage",
        **kwargs,
    ):
        select1_id = hg.html_id(self)
        select2_id = select1_id + "-n"
        kwargs["_class"] = kwargs.get("_class", "") + " bx--pagination"
        kwargs["data_pagination"] = True
        super().__init__(
            hg.DIV(
                hg.LABEL(
                    _("Items per page"),
                    ":",
                    _class="bx--pagination__text",
                    _for=select1_id,
                ),
                Select(
                    [
                        (
                            None,
                            [
                                {
                                    "label": i,
                                    "value": linkwithitemsperpage(
                                        itemsperpage_urlparameter, i, page_urlparameter
                                    ),
                                    "attrs": {
                                        "selected": hg.F(
                                            lambda c, i=i: c["request"].GET.get(
                                                itemsperpage_urlparameter
                                            )
                                            == str(i)
                                        )
                                    },
                                }
                                for i in items_per_page_options
                            ],
                        )
                    ],
                    inline=True,
                    widgetattributes={
                        "data_items_per_page": True,
                        "onchange": "document.location = this.value",
                        "onauxclick": "window.open(this.value, '_blank')",
                    },
                    _class="bx--select__item-count",
                ),
                hg.SPAN(
                    hg.SPAN(
                        hg.getattr_lazy(
                            get_page(paginator, page_urlparameter), "start_index"
                        ),
                        " - ",
                        hg.getattr_lazy(
                            get_page(paginator, page_urlparameter), "end_index"
                        ),
                        data_displayed_item_range=True,
                    ),
                    " ",
                    _("of"),
                    " ",
                    hg.SPAN(
                        " ",
                        hg.getattr_lazy(paginator, "count"),
                        " ",
                        data_total_items=True,
                    ),
                    " ",
                    _("items"),
                    _class="bx--pagination__text",
                ),
                _class="bx--pagination__left",
            ),
            hg.DIV(
                Select(
                    hg.F(
                        lambda c: [
                            (
                                None,
                                [
                                    {
                                        "label": i,
                                        "value": linktopage(page_urlparameter, i),
                                        "attrs": {
                                            "selected": hg.F(
                                                lambda c, i=i: c["request"].GET.get(
                                                    page_urlparameter, "1"
                                                )
                                                == str(i)
                                            ),
                                        },
                                    }
                                    for i in hg.resolve_lazy(paginator, c).page_range
                                ],
                            )
                        ]
                    ),
                    inline=True,
                    widgetattributes={
                        "data_page_number_input": True,
                        "onchange": "document.location = this.value",
                        "onauxclick": "window.open(this.value, '_blank')",
                    },
                    _class="bx--select__page-number",
                ),
                hg.LABEL(
                    _("of"),
                    " ",
                    hg.getattr_lazy(paginator, "num_pages"),
                    " ",
                    _("pages"),
                    _class="bx--pagination__text",
                    _for=select2_id,
                ),
                hg.BUTTON(
                    Icon("caret--left", size=20, _class="bx--pagination__nav-arrow"),
                    _class="bx--pagination__button bx--pagination__button--backward",
                    tabindex="0",
                    type="button",
                    disabled=hg.F(
                        lambda c: not paginator.get_page(
                            int(c["request"].GET.get(page_urlparameter, "1"))
                        ).has_previous()
                    ),
                    data_page_backward=True,
                    **aslink_attributes(
                        linktorelativepage(
                            page_urlparameter,
                            -1,
                            hg.getattr_lazy(paginator, "num_pages"),
                        )
                    ),
                ),
                hg.BUTTON(
                    Icon("caret--right", size=20, _class="bx--pagination__nav-arrow"),
                    _class="bx--pagination__button bx--pagination__button--forward",
                    tabindex="0",
                    type="button",
                    disabled=hg.F(
                        lambda c: not paginator.get_page(
                            int(c["request"].GET.get(page_urlparameter, "1"))
                        ).has_next()
                    ),
                    data_page_forward=True,
                    **aslink_attributes(
                        linktorelativepage(
                            page_urlparameter,
                            1,
                            hg.getattr_lazy(paginator, "num_pages"),
                        )
                    ),
                ),
                _class="bx--pagination__right",
            ),
            **kwargs,
        )

    @classmethod
    def from_config(cls, pagination_config):
        return cls(
            paginator=pagination_config.paginator,
            items_per_page_options=pagination_config.items_per_page_options,
            page_urlparameter=pagination_config.page_urlparameter,
            itemsperpage_urlparameter=pagination_config.itemsperpage_urlparameter,
        )


def linktopage(page_urlparameter, page):
    return hg.F(
        lambda c: link_with_urlparameters(
            c["request"], **{page_urlparameter: hg.resolve_lazy(page, c)}
        )
    )


def linktorelativepage(page_urlparameter, direction, maxnum):
    """direction: number of pages to jump, e.g. -1 or 1"""
    return hg.F(
        lambda c: link_with_urlparameters(
            c["request"],
            **{
                page_urlparameter: int(c["request"].GET.get(page_urlparameter, "1"))
                + hg.resolve_lazy(direction, c)
            },
        )
    )


def linkwithitemsperpage(itemsperpage_urlparameter, itemsperpage, page_urlparameter):
    return hg.F(
        lambda c: link_with_urlparameters(
            c["request"],
            **{itemsperpage_urlparameter: itemsperpage, page_urlparameter: None},
        )
    )


def get_page(paginator, page_urlparameter):
    def wrapper(context):
        return paginator.get_page(
            int(context["request"].GET.get(page_urlparameter, "1"))
        )

    return hg.F(wrapper)



import htmlgenerator as hg

from .helpers import ErrorListElement, HelpTextElement, LabelElement
from .icon import Icon
from .tag import Tag


class MultiSelect(hg.DIV):
    def __init__(
        self,
        optgroups,
        label=None,
        help_text=None,
        errors=None,
        disabled=None,
        required=None,
        # light=False, # TODO?
        # inline=False, # TODO?
        **attributes,
    ):
        """
        optgroups: list of 2-tuples: (group_name, group_items)
                   group_items: list of objects with attributes ['label', 'value', 'attrs'], attrs is a dict with html attributes

        """

        def countselected(context):
            options = [o for og in hg.resolve_lazy(optgroups, context) for o in og[1]]
            return len([o for o in options if o and o["selected"]])

        searchfieldid = hg.html_id(self)
        super().__init__(
            LabelElement(
                label,
                _for=searchfieldid,
                required=required,
                disabled=disabled,
            ),
            hg.If(
                disabled,
                hg.DIV(
                    hg.Iterator(
                        optgroups,
                        "optiongroup",
                        hg.Iterator(
                            hg.C("optiongroup.1"),
                            "option",
                            hg.If(hg.C("option.selected"), Tag(hg.C("option.label"))),
                        ),
                    )
                ),
                hg.DIV(
                    hg.DIV(
                        hg.If(
                            errors,
                            Icon(
                                "warning--filled",
                                size=16,
                                _class="bx--list-box__invalid-icon",
                            ),
                        ),
                        hg.DIV(
                            hg.F(countselected),
                            Icon(
                                "close",
                                focusable="false",
                                size=15,
                                role="img",
                                onclick="clearMultiselect(this.parentElement.parentElement.parentElement)",
                            ),
                            role="button",
                            _class="bx--list-box__selection bx--list-box__selection--multi bx--tag--filter",
                            tabindex="0",
                            title="Clear all selected items",
                        ),
                        hg.INPUT(
                            id=searchfieldid,
                            _class="bx--text-input",
                            placeholder="Filter...",
                            onclick="this.parentElement.nextElementSibling.style.display = 'block'",
                            onkeyup="filterOptions(this.parentElement.parentElement)",
                        ),
                        hg.DIV(
                            Icon(
                                "chevron--down", size=16, role="img", focusable="false"
                            ),
                            _class="bx--list-box__menu-icon",
                            onclick="this.parentElement.nextElementSibling.style.display = this.parentElement.nextElementSibling.style.display == 'none' ? 'block' : 'none';",
                        ),
                        role="button",
                        _class="bx--list-box__field",
                        tabindex="0",
                        onload="window.addEventListener('click', (e) => {this.nextElementSibling.style.display = 'none'})",
                    ),
                    hg.FIELDSET(
                        hg.LEGEND(
                            "Description of form elements within the fieldset",
                            _class="bx--assistive-text",
                        ),
                        hg.Iterator(
                            optgroups,
                            "optgroup",
                            hg.Iterator(
                                hg.C("optgroup.1"),
                                "option",
                                hg.DIV(
                                    hg.DIV(
                                        hg.DIV(
                                            hg.LABEL(
                                                hg.INPUT(
                                                    type="checkbox",
                                                    readonly=True,
                                                    _class="bx--checkbox",
                                                    value=hg.C("option.value"),
                                                    lazy_attributes=hg.C(
                                                        "option.attrs"
                                                    ),
                                                    onchange="updateMultiselect(this.closest('.bx--multi-select'))",
                                                    checked=hg.C("option.selected"),
                                                    name=hg.C("option.name"),
                                                ),
                                                hg.SPAN(
                                                    _class="bx--checkbox-appearance"
                                                ),
                                                hg.SPAN(
                                                    hg.C("option.label"),
                                                    _class="bx--checkbox-label-text",
                                                ),
                                                title=hg.C("option.label"),
                                                _class="bx--checkbox-label",
                                            ),
                                            _class="bx--form-item bx--checkbox-wrapper",
                                        ),
                                        _class="bx--list-box__menu-item__option",
                                    ),
                                    _class="bx--list-box__menu-item",
                                ),
                            ),
                        ),
                        _class="bx--list-box__menu",
                        role="listbox",
                        style="display: none",
                    ),
                    _class=hg.BaseElement(
                        "bx--multi-select bx--list-box bx--multi-select--selected bx--combo-box bx--multi-select--filterable",
                        hg.If(disabled, " bx--list-box--disabled"),
                    ),
                    data_invalid=hg.If(errors, True),
                ),
            ),
            HelpTextElement(help_text),
            ErrorListElement(errors),
            _class="bx--list-box__wrapper",
            onclick="event.stopPropagation()",
            **attributes,
        )



import htmlgenerator as hg

from .search import Search
from .tag import Tag


class SearchSelect(hg.DIV):
    def __init__(
        self,
        backend,
        boundfield,
        widgetattributes,
        **elementattributes,
    ):
        """
        :param SearchBackendConfig backend: Where and how to get search results
        """

        widgetattributes["value"] = widgetattributes["value"][0]
        # This works inside a formset. Might need to be changed for other usages.
        current_selection = getattr(
            boundfield.form.instance, elementattributes["fieldname"], ""
        )

        resultcontainerid = f"search-result-{widgetattributes['id']}"
        widget_id = widgetattributes["id"]
        tag_id = f"{widget_id}-tag"
        super().__init__(
            Tag(
                current_selection,
                id=tag_id,
                style=hg.If(
                    widgetattributes["value"] == "",
                    hg.BaseElement("display: none;"),
                ),
                onclick="return false;",
            ),
            hg.INPUT(_type="hidden", **widgetattributes),  # the actual form field
            Search(
                backend=backend,
                resultcontainerid=resultcontainerid,
                resultcontainer_onload_js=_resultcontainer_onload_js(
                    backend, resultcontainerid, tag_id, widget_id
                ),
                size="lg",
                disabled=widgetattributes.get("disabled", False),
                widgetattributes={"id": f"search__{widget_id}"},
            ),
            style="display: flex;",
            **elementattributes,
        )


def _resultcontainer_onload_js(backend, resultcontainerid, tag_id, widget_id):
    on_click = f"""function(evt) {{
        let label = $('{backend.result_label_selector}', this).innerHTML;
        let value = $('{backend.result_value_selector}', this).innerHTML;
        $('#{widget_id}').value = value;
        $('#{tag_id}').innerHTML = label;
        $('#{tag_id}').style = 'display: inline-block;';
        }}"""

    return f"""
    document.addEventListener('click', (evt) => this.innerHTML='');
    htmx.onLoad(function(target) {{
    // remove existing onlick attribute in case it e.g. redirects to the selected person
    $$('#{resultcontainerid} {backend.result_selector}')._.setAttribute('onclick', null);
    $$('#{resultcontainerid} {backend.result_selector}')._
    .addEventListener('click', {on_click});
    }});"""



from django.utils import formats

PHP_FORMAT_CHARACTERS = "dDjlNSwzWFmMntLoYyaABgGhHisuveIOPTZcrU"

# maps python format characters to PHP format characters
# hint: the #-sign of the python formatters is ommited for the keys
LETTER_MAPPING = {
    "a": "D",
    "A": "l",
    "w": "w",
    "d": "d",
    "b": "M",
    "B": "F",
    "m": "m",
    "y": "y",
    "Y": "Y",
    "H": "H",
    "I": "h",
    "p": "A",
    "M": "i",
    "S": "s",
    "f": "u",
    "z": "O",
    "Z": "e",
    "%": "%",
    "G": "o",
    "u": "N",
    "V": "W",
}


def to_php_formatstr(formatstr, format_key=None):
    """Maps a python datetime format string to a PHP format string

    This function is usefull because a lot of template/front-end code uses
    the PHP-format string, e.g. the date-filter of django.
    The following format specifiers are currently not supported:
    "%j", "%U", "%W", "%c", "%x", "%X"
    If formatstr is none, the format_key will be used to lookup default django format strings.
    """
    formatstr = " " + (formatstr or formats.get_format(format_key)[0])
    ret = []
    for i, c in enumerate(formatstr):
        if c == "%" or i == 0:
            continue

        if formatstr[i - 1] == "%":
            if c not in LETTER_MAPPING:
                raise ValueError(
                    f"Format letter %{c}  in format string {formatstr[1:]} is currently not supported to be converted to a PHP date-formatting string"
                )
            ret.append(LETTER_MAPPING[c])
        else:
            if c in PHP_FORMAT_CHARACTERS:
                ret.append("\\" + c)
            else:
                ret.append(c)
    return "".join(ret)



import htmlgenerator as hg

from . import button

REQUIRED_LABEL = " *"


class SubmitButton(hg.DIV):
    def __init__(self, *args, **kwargs):
        kwargs["type"] = "submit"
        super().__init__(button.Button(*args, **kwargs), _class="bx--form-item")


# TODO: A more generic approach for the form elements needs to be done
# I think we should replace the ErrorList, HelperText, Label things with the "smarter"
# version LabelElement etc. and add shortcuts to generate them directly from a field
# in the context instead of setting all the values inside the render-methods of the
# widgets. Or something like that. The API for form fields/widgets is currently not
# very great. For a test-implementation see text_input.py


class ErrorList(hg.DIV):
    def __init__(self, errors):
        super().__init__(
            hg.UL(*[hg.LI(e) for e in errors]),
            _class="bx--form-requirement",
        )


class HelperText(hg.DIV):
    def __init__(self, helpertext):
        super().__init__(helpertext, _class="bx--form__helper-text")


class Label(hg.LABEL):
    def __init__(self, *args, **kwargs):
        kwargs["_class"] = hg.BaseElement(kwargs.get("_class", ""), " bx--label")
        super().__init__(*args, **kwargs)


class LabelElement(hg.If):
    def __init__(self, label, _for, required=None, disabled=None, **kwargs):
        super().__init__(
            label,
            hg.LABEL(
                label,
                hg.If(required, REQUIRED_LABEL),
                _for=_for,
                _class=hg.BaseElement(
                    "bx--label",
                    hg.If(disabled, " bx--label--disabled"),
                ),
                **kwargs
            ),
        )


class HelpTextElement(hg.If):
    def __init__(self, helptext, disabled=False):
        super().__init__(
            helptext,
            hg.DIV(
                helptext,
                _class=hg.BaseElement(
                    "bx--form__helper-text",
                    hg.If(disabled, " bx--form__helper-text--disabled"),
                ),
            ),
        )


class ErrorListElement(hg.If):
    def __init__(self, errors):
        super().__init__(
            errors,
            hg.DIV(
                hg.UL(hg.Iterator(errors or (), "error", hg.C("error"))),
                _class="bx--form-requirement",
            ),
        )



import htmlgenerator as hg

from .helpers import REQUIRED_LABEL, ErrorListElement, HelpTextElement


class Checkbox(hg.DIV):
    def __init__(
        self,
        label=None,
        help_text=None,
        errors=None,
        disabled=None,
        required=None,
        widgetattributes={},
        **attributes,
    ):
        attributes["_class"] = (
            attributes.get("_class", "") + " bx--form-item bx--checkbox-wrapper"
        )
        widgetattributes["_class"] = (
            widgetattributes.get("_class", "") + " bx--checkbox"
        )
        widgetattributes["type"] = "checkbox"
        if disabled:
            widgetattributes["disabled"] = True
        self.input = hg.INPUT(**widgetattributes)
        self.label = hg.LABEL(
            self.input,
            label,
            hg.If(required, REQUIRED_LABEL),
            _class=hg.BaseElement(
                "bx--checkbox-label",
                hg.If(disabled, " bx--label--disabled"),
            ),
            data_contained_checkbox_state=hg.If(
                hg.F(lambda c: widgetattributes.get("checked", False)), "true", "false"
            ),
        )
        super().__init__(
            self.label,
            HelpTextElement(help_text),
            ErrorListElement(errors),
            **attributes,
        )



import datetime
import numbers
from collections.abc import Iterable

import htmlgenerator as hg
from dateutil import tz
from django.conf import settings
from django.db import models
from django.urls import NoReverseMatch
from django.utils.functional import Promise
from django.utils.html import format_html, format_html_join, linebreaks, mark_safe
from django_countries.fields import CountryField

import bread.settings as app_settings

from . import layout
from .utils.model_helpers import get_concrete_instance
from .utils.urls import reverse_model


def format_value(value, fieldtype=None):
    """Renders a python value in a nice way in HTML. If a field-definition has an attribute "renderer" set, that function will be used to render the value"""
    if hasattr(fieldtype, "renderer"):
        return fieldtype.renderer(value)

    if isinstance(value, bool) or value is None:
        return CONSTANTS[value]

    # make referencing fields iterable (normaly RelatedManagers)
    if isinstance(value, models.Manager):
        value = value.all()

    # If there is a hint passed via fieldtype, use the accoring conversion function first (identity otherwise)
    # This is mostly helpfull for string-based fields like URLS, emails etc.
    value = MODELFIELD_FORMATING_HELPERS.get(fieldtype.__class__, lambda a: a)(value)

    if isinstance(value, bool) or value is None:
        return CONSTANTS[value]
    if isinstance(value, datetime.timedelta):
        return as_duration(value)
    if isinstance(value, datetime.datetime):
        return as_datetime(value)
    if isinstance(value, numbers.Number):
        return f"{float(value):f}".rstrip("0").rstrip(".")
    if isinstance(value, models.fields.files.ImageFieldFile):
        return as_image(value)
    if isinstance(value, models.fields.files.FieldFile):
        return as_download(value)
    if isinstance(value, Iterable) and not isinstance(value, (str, bytes, Promise)):
        return as_list(value)
    if isinstance(value, models.Model):
        try:
            return as_object_link(value)
        except NoReverseMatch:
            return value
    return value


# Formatting functions: never pass None, always return string


def as_email(value):
    return format_html('<a href="mailto:{}">{}</a>', value, value)


def as_url(value):
    return format_html(
        '<a href="{}" target="_blank" rel="noopener noreferrer">{}</a>', value, value
    )


def as_text(value):
    return mark_safe(linebreaks(value, autoescape=True))


def as_time(value):
    if value.tzinfo:
        value = value.astimezone(tz.gettz(settings.TIME_ZONE))
    return f"{value.hour:02}:{value.minute:02}:{value.second:02}"


def as_duration(value):
    return str(value - datetime.timedelta(microseconds=value.microseconds))


def as_datetime(value):
    if value.tzinfo:
        value = value.astimezone(tz.gettz(settings.TIME_ZONE))

    return value.isoformat(sep=" ", timespec="seconds").rsplit("+", 1)[0]


def as_countries(value):
    return as_list((c.name for c in value))


def as_list(iterable, sep=", "):
    return format_html_join(sep, "{}", ((format_value(v),) for v in iterable))


def as_richtext(value):
    return mark_safe(value)


def as_download(value):
    if not value:
        return CONSTANTS[None]
    if not value.storage.exists(value.name):
        return mark_safe("<small><emph>File not found</emph></small>")
    return mark_safe(
        hg.render(
            hg.BaseElement(
                hg.A(
                    layout.icon.Icon("launch", size=16),
                    newtab=True,
                    href=value.url,
                    style="margin-right: 0.5rem; margin-left: 0.5rem",
                ),
                hg.A(
                    layout.icon.Icon("download", size=16), download=True, href=value.url
                ),
            ),
            {},
        )
    )


def as_image(value):
    if not value:
        return CONSTANTS[None]
    if not value.storage.exists(value.name):
        return mark_safe("<small><emph>Image not found</emph></small>")
    from easy_thumbnails.files import get_thumbnailer

    im = get_thumbnailer(value).get_thumbnail({"size": (100, 100), "quality": 75})
    return format_html(
        '<a class="center" href="{}"><img src={} width="{}" height="{}"/></a>',
        value.url,
        im.url,
        im.width,
        im.height,
    )


def as_audio(value):
    if not value:
        return CONSTANTS[None]
    if not value.storage.exists(value.name):
        return mark_safe("<small><emph>Audio file not found</emph></small>")
    return format_html(
        """
        <audio controls controlsList="nodownload" preload="metadata">
            <source src="{}" type="audio/mp3">
        </audio>
    """,
        value.url,
    )


def as_video(value):
    if not value:
        return CONSTANTS[None]
    if not value.storage.exists(value.name):
        return mark_safe("<small><emph>Video file not found</emph></small>")
    return format_html(
        """
        <video controls width="320" height="240" controlsList="nodownload" preload="metadata">
            <source src="{}" type="video/mp4">
        </video>
    """,
        value.url,
    )


def as_object_link(obj, label=None):
    if hasattr(obj, "get_absolute_url"):
        return format_html('<a href="{}">{}</a>', obj.get_absolute_url(), obj)

    return format_html(
        '<a href="{}">{}</a>',
        reverse_model(get_concrete_instance(obj), "read"),
        label or obj.__str__(),
    )


MODELFIELD_FORMATING_HELPERS = {
    None: lambda a: a,
    models.EmailField: as_email,
    models.ImageField: as_image,
    models.FileField: as_download,
    models.URLField: as_url,
    models.TextField: as_text,
    models.TimeField: as_time,
    models.DateTimeField: as_datetime,
    CountryField: as_countries,
}

CONSTANTS = {
    None: getattr(settings, "HTML_NONE", app_settings.HTML_NONE),
    True: getattr(settings, "HTML_TRUE", app_settings.HTML_TRUE),
    False: getattr(settings, "HTML_FALSE", app_settings.HTML_FALSE),
}



import subprocess  # nosec because we covered everything

import htmlgenerator as hg
import pkg_resources
from django.utils.translation import gettext_lazy as _

from ..utils import aslayout


@aslayout
def systeminformation(request):
    git_status = ""
    try:
        git_status = (
            subprocess.run(  # nosec because we have no user input to subprocess
                ["git", "log", "-n", "5", "--oneline"], capture_output=True, check=True
            ).stdout.decode()
        )
    except subprocess.SubprocessError as e:
        git_status = hg.BaseElement(
            "ERROR",
            hg.BR(),
            str(e),
            hg.BR(),
            getattr(e, "stdout", b"").decode(),
            hg.BR(),
            getattr(e, "stderr", b"").decode(),
        )

    return hg.BaseElement(
        hg.H3(_("System information")),
        hg.H4("Git log"),
        hg.PRE(hg.CODE(git_status)),
        hg.H4("PIP packages", style="margin-top: 2rem"),
        hg.UL(
            hg.Iterator(
                sorted(
                    ["%s==%s" % (i.key, i.version) for i in pkg_resources.working_set]
                ),
                "package",
                hg.LI(hg.C("package")),
            )
        ),
    )



import json

import htmlgenerator as hg
from django.core import checks
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.urls import reverse
from django.utils.html import mark_safe
from django.utils.translation import gettext_lazy as _
from djangoql.exceptions import DjangoQLError
from djangoql.queryset import apply_search
from djangoql.schema import DjangoQLSchema
from djangoql.serializers import DjangoQLSchemaSerializer

from bread import layout


class QueryValue:
    def __init__(self, queryset, raw):
        self.queryset = queryset
        self.raw = raw

    def __str__(self):
        return self.raw


class QuerySetDescriptor:
    def __init__(self, field):
        self.field = field

    def __get__(self, instance=None, owner=None):
        if instance is None:
            return self
        if self.field.name not in instance.__dict__:
            instance.refresh_from_db(fields=[self.field.name])
        value = instance.__dict__[self.field.name]
        model = getattr(instance, self.field.modelfieldname, None)
        if model and model.model_class():
            try:
                return parsequeryexpression(model.model_class().objects, value)
            except ValidationError:
                pass
        return QueryValue(None, value)

    def __set__(self, instance, value):
        instance.__dict__[self.field.name] = self.field.get_clean_value(value)


class QuerysetField(models.TextField):
    descriptor_class = QuerySetDescriptor

    def __init__(self, *args, modelfieldname, **kwargs):
        self.modelfieldname = modelfieldname
        kwargs["blank"] = True
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        return name, path, args, {"modelfieldname": self.modelfieldname, **kwargs}

    def contribute_to_class(self, cls, name):
        super().contribute_to_class(cls, name)
        setattr(cls, self.name, self.descriptor_class(self))

    def get_clean_value(self, value):
        return str(value)

    def get_prep_value(self, value):
        super().get_prep_value(self.get_clean_value(value))

    def value_to_string(self, obj):
        return self.get_prep_value(self.value_from_object(obj))

    def get_db_prep_value(self, value, connection, prepared=False):
        return self.get_clean_value(value)

    def formfield(self, **kwargs):
        ret = super().formfield(**kwargs)
        ret.layout = QuerySetFormWidget
        ret.layout_kwargs = {
            "modelfieldname": self.modelfieldname,
            "rows": 1,
            "name": self.name,
        }
        return ret

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_field_name(),
        ]

    def _check_field_name(self):
        try:
            self.model._meta.get_field(self.modelfieldname)
        except FieldDoesNotExist:
            return [
                checks.Error(
                    "The QuerysetField modelfieldname references the "
                    "nonexistent field '%s'." % self.modelfieldname,
                    obj=self,
                    id="queryfield.E001",
                )
            ]
        else:
            return []

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        model = getattr(model_instance, self.modelfieldname)
        if not model or not model.model_class():
            raise ValidationError(_("Invalid model '%s' selected") % model)
        parsequeryexpression(
            getattr(model_instance, self.modelfieldname).model_class().objects, value
        )


class QuerySetFormWidget(layout.text_area.TextArea):
    def __init__(self, *args, modelfieldname, name, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.boundfield = kwargs.get("boundfield", None)
        self.model = None
        if "boundfield" in kwargs:
            self.model = getattr(
                self.boundfield.form.instance, modelfieldname
            ).model_class()

    def render(self, context):
        if self.model:
            self.append(
                hg.SCRIPT(
                    mark_safe(
                        """
    document.addEventListener("DOMContentLoaded", () => DjangoQL.DOMReady(function () {
    new DjangoQL({
        introspections: %s,
        selector: 'textarea[name=%s]',
        syntaxHelp: '%s',
        autoResize: false
      });
    }));
    """
                        % (
                            json.dumps(
                                DjangoQLSchemaSerializer().serialize(
                                    DjangoQLSchema(self.model)
                                )
                            ),
                            self.name,
                            reverse("reporthelp"),
                        )
                    )
                ),
            )
        return super().render(context)


def parsequeryexpression(basequeryset, expression):
    if not expression:
        return QueryValue(basequeryset.all(), expression)
    try:
        return QueryValue(apply_search(basequeryset, expression), expression)
    except DjangoQLError as e:
        raise ValidationError(str(e))



# Generated by Django 3.1.7 on 2021-06-19 08:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0003_report_custom_queryset'),
    ]

    operations = [
        migrations.AlterField(
            model_name='report',
            name='custom_queryset',
            field=models.CharField(blank=True, help_text="Key in 'settings.REPORT_FILTERS' must be a function returning a queryset", max_length=255, verbose_name='Custom Filter'),
        ),
    ]



import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from .icon import Icon


class Button(hg.BUTTON):
    """buttontype: "primary", "secondary", "tertiary", "danger", "ghost" """

    def __init__(
        self,
        *children,
        buttontype="primary",
        icon=None,
        notext=False,
        small=False,
        **attributes,
    ):
        attributes["type"] = attributes.get("type", "button")
        attributes["tabindex"] = attributes.get("tabindex", "0")
        attributes["_class"] = (
            attributes.get("_class", "") + f" bx--btn bx--btn--{buttontype}"
        )
        if small:
            attributes["_class"] += " bx--btn--sm "
        if notext or not children:
            attributes["_class"] += " bx--btn--icon-only"
            if children:
                attributes[
                    "_class"
                ] += " bx--btn--icon-only bx--tooltip__trigger bx--tooltip--a11y bx--tooltip--bottom bx--tooltip--align-center"
                children = (hg.SPAN(*children, _class="bx--assistive-text"),)

        if icon is not None:
            if isinstance(icon, str):
                icon = Icon(icon)
            icon.attributes["_class"] = (
                icon.attributes.get("_class", "") + " bx--btn__icon"
            )
            children += (icon,)
        super().__init__(*children, **attributes)

    @staticmethod
    def fromlink(link, **kwargs):
        buttonargs = {
            "icon": link.iconname,
            "notext": not link.label,
        }
        return Button(
            *([link.label] if link.label else []),
            **{**buttonargs, **link.attributes, **kwargs},
        ).as_href(link.href)

    def as_href(self, href):
        return hg.A(*self, **{**self.attributes, "href": href})


class ButtonSet(hg.htmltags.DIV):
    def __init__(self, *buttons, **attributes):
        attributes["_class"] = attributes.get("_class", "") + " bx--btn-set"
        super().__init__(*buttons, **attributes)


class PrintPageButton(Button):
    def __init__(self, **attributes):
        if "onclick" not in attributes:
            attributes["onclick"] = "window.print()"
        super().__init__(_("Print"), icon="printer", notext=True, **attributes)



import htmlgenerator as hg
from django.conf import settings
from django.http import HttpResponse
from django.template.context import _builtin_context_processors
from django.utils.formats import localize
from django.utils.module_loading import import_string

from bread.utils import pretty_modelname, resolve_modellookup
from bread.utils.urls import reverse_model

from ..formatters import format_value

DEVMODE_KEY = "DEVMODE"


class HasBreadCookieValue(hg.Lazy):
    def __init__(self, cookiename, value, default=None):
        self.cookiename = cookiename
        self.value = value
        self.default = default

    def resolve(self, context):
        if f"bread-{self.cookiename}" in context["request"].session.get(
            "bread-cookies", {}
        ):
            return (
                context["request"].session["bread-cookies"][f"bread-{self.cookiename}"]
                == self.value
            )
        return self.default == self.value


class DevModeOnly(hg.BaseElement):
    def render(self, context):
        if context["request"].session.get(DEVMODE_KEY, False):
            return super().render(context)
        return ""


def fieldlabel(model, accessor):
    label = resolve_modellookup(model, accessor)[-1]
    if hasattr(label, "verbose_name"):
        return label.verbose_name
    if isinstance(label, property):
        return label.fget.__name__.replace("_", " ")
    if callable(label):
        return label.__name__.replace("_", " ")
    return label


def objectaction(object, action, *args, **kwargs):
    kwargs["kwargs"] = {"pk": object.pk}
    return str(
        reverse_model(
            object,
            action,
            *args,
            **kwargs,
        )
    )


def aslink_attributes(href):
    """
    Shortcut to generate HTMLElement attributes to make any element behave like a link.
    This should normally be used like this: hg.DIV("hello", \\*\\*aslink_attributes('google.com'))
    """
    return {
        "onclick": hg.BaseElement("document.location = '", href, "'"),
        "onauxclick": hg.BaseElement("window.open('", href, "', '_blank')"),
        "style": "cursor: pointer",
    }


class ModelName(hg.ContextValue):
    def resolve(self, context):
        return pretty_modelname(super().resolve(context))


class FormattedContextValue(hg.ContextValue):
    def resolve(self, context):
        return str(format_value(super().resolve(context)))


class ObjectFieldLabel(hg.ContextValue):
    def __init__(self, fieldname, object_contextname="object", title=True):
        """
        :param fieldname: Name of the model field whose value will be rendered
        :param object_contextname: Name of the context object which provides the field value
        """
        super().__init__(object_contextname)
        self.fieldname = fieldname
        self.title = title

    def resolve(self, context):
        ret = fieldlabel(super().resolve(context)._meta.model, self.fieldname)
        return ret.title() if self.title and isinstance(ret, str) else ret


class ObjectFieldValue(hg.ContextValue):
    def __init__(self, fieldname, object_contextname="object", formatter=None):
        """
        :param fieldname: Name of the model field whose value will be rendered
        :param object_contextname: Name of the context object which provides the field value
        :param formatter: function which takes the field value as a single argument and returns a formatted version
        """
        super().__init__(object_contextname)
        self.fieldname = fieldname
        self.formatter = formatter

    def resolve(self, context):
        object = super().resolve(context)
        parts = self.fieldname.split(".")
        # test if the value has a matching get_FIELDNAME_display function
        value = hg.resolve_lookup(
            object, f"{'.'.join(parts[:-1])}.get_{parts[-1]}_display"
        )
        if value is None:
            value = hg.resolve_lookup(object, self.fieldname)
        return (
            self.formatter(value)
            if self.formatter
            else localize(value, use_l10n=settings.USE_L10N)
        )


FC = FormattedContextValue


def render(request, layout, context=None, **response_kwargs):
    if render.CONTEXT_PROCESSORS is None:
        render.CONTEXT_PROCESSORS = tuple(
            import_string(path)
            for path in _builtin_context_processors
            + tuple(
                (settings.TEMPLATES + [{}])[0]
                .get("OPTIONS", {})
                .get("context_processors", [])
            )
        )
    response_kwargs.setdefault("content_type", "text/html")
    defaultcontext = {}
    for processor in render.CONTEXT_PROCESSORS:
        defaultcontext.update(processor(request))
    defaultcontext.update(context or {})
    return HttpResponse(layout.render(defaultcontext), **response_kwargs)


render.CONTEXT_PROCESSORS = None



from typing import Any, Iterable, List, NamedTuple, Optional, Union

import htmlgenerator as hg
from django.db import models
from django.utils.translation import gettext_lazy as _

from bread.utils import filter_fieldlist, pretty_modelname, resolve_modellookup
from bread.utils.links import Link, ModelHref
from bread.utils.urls import link_with_urlparameters

from ..base import ObjectFieldLabel, ObjectFieldValue, aslink_attributes
from .button import Button
from .icon import Icon
from .overflow_menu import OverflowMenu
from .pagination import Pagination, PaginationConfig
from .search import Search, SearchBackendConfig


class DataTableColumn(NamedTuple):
    header: Any
    cell: Any
    sortingname: Optional[str] = None
    enable_row_click: bool = True
    th_attributes: hg.F = None
    td_attributes: hg.F = None

    @staticmethod
    def from_modelfield(
        col,
        model,
        prevent_automatic_sortingnames=False,
        rowvariable="row",
        th_attributes=None,
        td_attributes=None,
    ) -> "DataTableColumn":
        return DataTableColumn(
            ObjectFieldLabel(col, model),
            ObjectFieldValue(col, rowvariable),
            sortingname_for_column(model, col)
            if not prevent_automatic_sortingnames
            else None,
            th_attributes=th_attributes,
            td_attributes=td_attributes,
        )

    def as_header_cell(self, orderingurlparameter="ordering"):
        headcontent = hg.SPAN(self.header, _class="bx--table-header-label")
        if self.sortingname:
            headcontent = hg.BUTTON(
                headcontent,
                Icon("arrow--down", _class="bx--table-sort__icon", size=16),
                Icon(
                    "arrows--vertical",
                    _class="bx--table-sort__icon-unsorted",
                    size=16,
                ),
                _class=hg.BaseElement(
                    "bx--table-sort ",
                    sortingclass_for_column(orderingurlparameter, self.sortingname),
                ),
                data_event="sort",
                title=self.header,
                **sortinglink_for_column(orderingurlparameter, self.sortingname),
            )
        return hg.TH(headcontent, lazy_attributes=self.th_attributes)


class DataTable(hg.TABLE):
    SPACINGS = ["default", "compact", "short", "tall"]

    def __init__(
        self,
        columns: List[DataTableColumn],
        row_iterator: Union[hg.Lazy, Iterable, hg.Iterator],
        orderingurlparameter: str = "ordering",
        rowvariable: str = "row",
        spacing: str = "default",
        zebra: bool = False,
        **kwargs: dict,
    ):
        """A carbon DataTable element

        :param columns: Column definitions
        :param row_iterator: Iterator which yields row objects. If this is a hg.Iterator instance it will be used for the table body, otherwise a default Iterator will be used to render the column cells. This can also be htmlgenerator.Lazy object which returns a Python iterator when beeing evaluated.
        :param rowvariable: Name of the current object passed to childrens context
        :param orderingurlparameter: The name of the GET query parameter which is used to set the table ordering
        :param spacing: One of "default", "compact", "short", "tall", according to the carbon styles
        :param zebra: If True alternate row colors
        :param kwargs: HTML element attributes
        """

        self.head = DataTable.headrow(columns, orderingurlparameter)
        if isinstance(row_iterator, hg.Iterator):
            self.iterator = row_iterator
        else:
            self.iterator = hg.Iterator(
                row_iterator, rowvariable, DataTable.row(columns)
            )
        kwargs["_class"] = kwargs.get("_class", "") + " ".join(
            DataTable.tableclasses(spacing, zebra)
        )
        super().__init__(hg.THEAD(self.head), hg.TBODY(self.iterator), **kwargs)

    def with_toolbar(
        self,
        title: Any,
        helper_text: Any = None,
        primary_button: Optional[Button] = None,
        search_backend: Optional[SearchBackendConfig] = None,
        bulkactions: List[Link] = (),
        pagination_config: Optional[PaginationConfig] = None,
        checkbox_for_bulkaction_name="_selected",
        settingspanel: Any = None,
    ):
        """
        wrap this datatable with title and toolbar
        title: table title
        helper_text: sub title
        primary_button: bread.layout.button.Button instance
        search_backend: search endpoint
        bulkactions: List of bread.utils.links.Link instances. Will send a post or a get (depending on its "method" attribute) to the target url the sent data will be a form with the selected checkboxes as fields if the head-checkbox has been selected only that field will be selected
        """
        checkboxallid = f"datatable-check-{hg.html_id(self)}"
        header = [hg.H4(title, _class="bx--data-table-header__title")]
        if helper_text is not None:
            header.append(
                hg.P(helper_text, _class="bx--data-table-header__description")
            )
        resultcontainerid = f"datatable-search-{hg.html_id(self)}"
        bulkactionlist = []
        for link in bulkactions:
            bulkactionlist.append(
                Button(
                    link.label,
                    icon=link.iconname,
                    onclick=hg.BaseElement(
                        "submitbulkaction(this.closest('[data-table]'), '",
                        link.href,
                        "', method='GET')",
                    ),
                ),
            )

        if bulkactions:
            self.head.insert(
                0,
                hg.TH(
                    hg.INPUT(
                        data_event="select-all",
                        id=checkboxallid,
                        _class="bx--checkbox",
                        type="checkbox",
                        name=checkbox_for_bulkaction_name,
                        value="all",
                    ),
                    hg.LABEL(
                        _for=checkboxallid,
                        _class="bx--checkbox-label",
                    ),
                    _class="bx--table-column-checkbox",
                ),
            )
            self.iterator[0].insert(
                0,
                hg.TD(
                    hg.INPUT(
                        data_event="select",
                        id=hg.BaseElement(
                            checkboxallid,
                            "-",
                            hg.C(self.iterator.loopvariable + "_index"),
                        ),
                        _class="bx--checkbox",
                        type="checkbox",
                        name=checkbox_for_bulkaction_name,
                        value=hg.If(
                            hg.F(
                                lambda c: hasattr(c[self.iterator.loopvariable], "pk")
                            ),
                            hg.C(f"{self.iterator.loopvariable}.pk"),
                            hg.C(f"{self.iterator.loopvariable}_index"),
                        ),
                    ),
                    hg.LABEL(
                        _for=hg.BaseElement(
                            checkboxallid,
                            "-",
                            hg.C(self.iterator.loopvariable + "_index"),
                        ),
                        _class="bx--checkbox-label",
                        aria_label="Label name",
                    ),
                    _class="bx--table-column-checkbox",
                ),
            )

        return hg.DIV(
            hg.DIV(*header, _class="bx--data-table-header"),
            hg.SECTION(
                hg.DIV(
                    hg.DIV(
                        *(
                            bulkactionlist
                            + [
                                Button(
                                    _("Cancel"),
                                    data_event="action-bar-cancel",
                                    _class="bx--batch-summary__cancel",
                                )
                            ]
                        ),
                        _class="bx--action-list",
                    ),
                    hg.DIV(
                        hg.P(
                            hg.SPAN(0, data_items_selected=True),
                            _(" items selected"),
                            _class="bx--batch-summary__para",
                        ),
                        _class="bx--batch-summary",
                    ),
                    _class="bx--batch-actions",
                    aria_label=_("Table Action Bar"),
                ),
                hg.DIV(
                    hg.DIV(
                        Search(
                            widgetattributes={"autofocus": True},
                            backend=search_backend,
                            resultcontainerid=resultcontainerid,
                            show_result_container=False,
                        ),
                        _class="bx--toolbar-search-container-persistent",
                    )
                    if search_backend
                    else None,
                    Button(
                        icon="settings--adjust",
                        buttontype="ghost",
                        onclick="this.parentElement.parentElement.parentElement.querySelector('.settingscontainer').style.display = this.parentElement.parentElement.parentElement.querySelector('.settingscontainer').style.display == 'block' ? 'none' : 'block'; event.stopPropagation()",
                    )
                    if settingspanel
                    else None,
                    primary_button or None,
                    _class="bx--toolbar-content",
                ),
                _class="bx--table-toolbar",
            ),
            hg.DIV(
                hg.DIV(
                    id=resultcontainerid,
                    style="width: 100%; position: absolute; z-index: 999",
                ),
                style="width: 100%; position: relative",
            ),
            hg.DIV(
                hg.DIV(
                    hg.DIV(
                        settingspanel,
                        _class="bx--tile raised",
                        style="margin: 0; padding: 0",
                    ),
                    _class="settingscontainer",
                    style="position: absolute; z-index: 999; right: 0; display: none",
                    onload="document.addEventListener('click', (e) => {this.style.display = 'none'})",
                ),
                style="position: relative",
                onclick="event.stopPropagation()",
            ),
            self,
            Pagination.from_config(pagination_config) if pagination_config else None,
            _class="bx--data-table-container",
            data_table=True,
        )

    @staticmethod
    def from_model(
        model,
        queryset=None,
        # column behaviour
        columns: List[Union[str, DataTableColumn]] = None,
        prevent_automatic_sortingnames=False,
        # row behaviour
        rowvariable="row",
        rowactions: List[Link] = (),
        rowactions_dropdown=False,
        rowclickaction=None,
        # bulkaction behaviour
        bulkactions: List[Link] = (),
        checkbox_for_bulkaction_name="_selected",
        # toolbar configuration
        title=None,
        primary_button: Optional[Button] = None,
        settingspanel: Any = None,
        search_backend: Optional[SearchBackendConfig] = None,
        pagination_config: Optional[PaginationConfig] = None,
        **kwargs,
    ):
        """TODO: Write Docs!!!!
        Yeah yeah, on it already...

        :param settingspanel: A panel which will be opened when clicking on the
                              "Settings" button of the datatable, usefull e.g.
                              for showing filter options. Currently only one
                              button and one panel are supported. More buttons
                              and panels could be interesting but may to over-
                              engineered because it is a rare case and it is not
                              difficutl to add another button by modifying the
                              datatable after creation.
        """
        for col in columns:
            if not (isinstance(col, DataTableColumn) or isinstance(col, str)):
                raise ValueError(
                    f"Argument 'columns' needs to be of a List[str] or a List[DataTableColumn], but found {col}"
                )

        title = title or pretty_modelname(model, plural=True)

        if primary_button is None:
            primary_button = Button.fromlink(
                Link(
                    href=ModelHref(model, "add"),
                    label=_("Add %s") % pretty_modelname(model),
                ),
                icon=Icon("add", size=20),
            )

        if rowactions_dropdown:
            objectactions_menu = OverflowMenu(
                rowactions,
                flip=True,
                item_attributes={"_class": "bx--table-row--menu-option"},
            )
        else:
            objectactions_menu = hg.DIV(
                hg.Iterator(
                    rowactions,
                    "action",
                    hg.F(
                        lambda c: Button.fromlink(
                            c["action"],
                            notext=True,
                            small=True,
                            buttontype="ghost",
                            _class="bx--overflow-menu",
                        )
                    ),
                ),
                style="display: flex; justify-content: flex-end;",
            )

        queryset = model.objects.all() if queryset is None else queryset
        columns = columns or filter_fieldlist(model, ["__all__"])
        column_definitions: List[DataTableColumn] = []
        for col in columns:
            td_attributes = None
            if rowclickaction and getattr(col, "enable_row_click", True):
                assert isinstance(
                    rowclickaction, Link
                ), "rowclickaction must be of type Link"
                td_attributes = {
                    **aslink_attributes(rowclickaction.href),
                    **(rowclickaction.attributes or {}),
                }
            # convert simple string (modelfield) to column definition
            if isinstance(col, str):
                col = DataTableColumn.from_modelfield(
                    col,
                    model,
                    prevent_automatic_sortingnames,
                    rowvariable,
                    td_attributes=td_attributes,
                )
            else:
                if td_attributes:
                    col = col._replace(td_attributes=td_attributes)

            column_definitions.append(col)

        return DataTable(
            column_definitions
            + (
                [
                    DataTableColumn(
                        "",
                        objectactions_menu,
                        td_attributes=hg.F(
                            lambda c: {
                                "_class": "bx--table-column-menu"
                                if rowactions_dropdown
                                else ""
                            }
                        ),
                        th_attributes=hg.F(
                            lambda c: {"_class": "bx--table-column-menu"}
                        ),
                    )
                ]
                if rowactions
                else []
            ),
            # querysets are cached, the call to all will make sure a new query is used in every request
            hg.F(lambda c: queryset),
            **kwargs,
        ).with_toolbar(
            title,
            helper_text=hg.format(
                "{} {}",
                hg.F(lambda c: len(hg.resolve_lazy(queryset, c))),
                model._meta.verbose_name_plural,
            ),
            primary_button=primary_button,
            search_backend=search_backend,
            bulkactions=bulkactions,
            pagination_config=pagination_config,
            checkbox_for_bulkaction_name=checkbox_for_bulkaction_name,
            settingspanel=settingspanel,
        )

    @staticmethod
    def from_queryset(
        queryset,
        **kwargs,
    ):
        return DataTable.from_model(
            queryset.model,
            queryset=queryset,
            **kwargs,
        )

    # A few helper classes to make the composition in the __init__ method easier

    @staticmethod
    def headrow(columns, orderingurlparameter):
        """Returns the head-row element based on the specified columns"""
        return hg.TR(*[c.as_header_cell(orderingurlparameter) for c in columns])

    @staticmethod
    def row(columns):
        """Returns a row element based on the specified columns"""
        return hg.TR(
            *[hg.TD(col.cell, lazy_attributes=col.td_attributes) for col in columns]
        )

    @staticmethod
    def tableclasses(spacing, zebra):
        if spacing not in DataTable.SPACINGS:
            raise ValueError(
                f"argument 'spacin' is {spacing} but needs to be one of {DataTable.SPACINGS}"
            )
        classes = ["bx--data-table bx--data-table--sort"]
        if spacing != "default":
            classes.append(f" bx--data-table--{spacing}")
        if zebra:
            classes.append(" bx--data-table--zebra")
        return classes


def sortingclass_for_column(orderingurlparameter, columnname):
    def extracturlparameter(context):
        value = context["request"].GET.get(orderingurlparameter, "")
        if not value:
            return ""
        if value == columnname:
            return "bx--table-sort--active"
        if value == "-" + columnname:
            return "bx--table-sort--active bx--table-sort--ascending"
        return ""

    return hg.F(extracturlparameter)


def sortingname_for_column(model, column):
    components = []
    for field in resolve_modellookup(model, column):
        if hasattr(field, "sorting_name"):
            components.append(field.sorting_name)
        elif isinstance(field, models.Field):
            components.append(field.name)
        else:
            return None
    return "__".join(components)


def sortinglink_for_column(orderingurlparameter, columnname):
    ORDERING_VALUES = {
        None: columnname,
        columnname: "-" + columnname,
        "-" + columnname: None,
    }

    def extractsortinglink(context):
        currentordering = context["request"].GET.get(orderingurlparameter, None)
        nextordering = ORDERING_VALUES.get(currentordering, columnname)
        return link_with_urlparameters(
            context["request"], **{orderingurlparameter: nextordering}
        )

    return aslink_attributes(hg.F(extractsortinglink))



# Generated by Django 3.1.7 on 2021-06-19 04:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reports', '0002_auto_20210322_1115'),
    ]

    operations = [
        migrations.AddField(
            model_name='report',
            name='custom_queryset',
            field=models.CharField(blank=True, help_text='python path to function which returns a queryset, ignores filter field', max_length=255, verbose_name='Custom Filter'),
        ),
    ]



from celery import Task, shared_task
from django.apps import AppConfig

from . import settings


class WorkflowsConfig(AppConfig):
    name = "bread.contrib.workflows"

    def ready(self):
        shared_task(base=WorkflowUpdate, run_every=settings.WORKFLOW_BEAT)(
            update_workflows
        )


class WorkflowUpdate(Task):
    @classmethod
    def on_bound(cls, app):
        app.conf.beat_schedule[cls.name] = {
            "task": cls.name,
            "schedule": cls.run_every.total_seconds(),
            "args": (),
            "kwargs": {},
            "options": getattr(cls, "options", {}),
            "relative": getattr(cls, "relative", False),
        }


def update_workflows():
    from .models import WorkflowBase

    for cls in WorkflowBase.__subclasses__():
        for workflow in cls.objects.filter(
            cancelled__isnull=True, completed__isnull=True
        ):
            workflow.save()



import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from .button import Button
from .helpers import ErrorListElement, HelpTextElement, LabelElement
from .icon import Icon


class TextInput(hg.DIV):
    def __init__(
        self,
        light=False,
        widgetattributes=None,
        label=None,
        help_text=None,
        errors=None,
        required=None,
        disabled=None,
        icon=None,
        **attributes,
    ):
        # Pop these attributes because "attributes" should be passed
        # as HTML attributes to the parent __init__ and they do not render
        # properly sometimes (e.g. boundfield will render the fields HTML
        # inside the attribute
        # TODO: Define and implement a better interface for components which can
        # be instantiated in bread.layout.component._mapwidget
        # See https://github.com/basxsoftwareassociation/bread/issues/16
        attributes.pop("boundfield")
        attributes.pop("fieldname")
        attributes["_class"] = (
            attributes.get("_class", "") + " bx--text-input-wrapper bx--form-item"
        )
        widgetattributes = widgetattributes or {}
        widgetattributes["_class"] = hg.BaseElement(
            widgetattributes.get("_class", ""),
            " bx--text-input",
            hg.If(light, " bx--text-input--light"),
            hg.If(errors, " bx--text-input--invalid"),
        )

        super().__init__(
            LabelElement(
                label,
                widgetattributes.get("id"),
                required=required,
                disabled=disabled,
            ),
            hg.DIV(
                hg.If(
                    errors,
                    Icon(
                        "warning--filled",
                        size=16,
                        _class="bx--text-input__invalid-icon",
                    ),
                ),
                hg.INPUT(
                    type="text",
                    **widgetattributes,
                ),
                hg.If(
                    icon,
                    Icon(
                        icon,
                        size=16,
                        _class="text-input-icon",
                    ),
                ),
                _class=(
                    "bx--text-input__field-wrapper"
                    + (" text-input-with-icon" if icon is not None else "")
                ),
                data_invalid=hg.If(errors, True),
            ),
            HelpTextElement(
                help_text, disabled=widgetattributes.get("disabled", False)
            ),
            ErrorListElement(errors),
            **attributes,
        )


class PasswordInput(TextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attributes["data-text-input"] = True
        self.attributes["_class"] += " bx--password-input-wrapper"
        inputElement = self[1][1]
        inputElement.attributes["type"] = "password"
        inputElement.attributes["data-toggle-password-visibility"] = True
        inputElement.attributes["_class"] += " bx--password-input"
        showhidebtn = Button(_("Show password"), notext=True)
        showhidebtn.attributes[
            "_class"
        ] = "bx--text-input--password__visibility__toggle bx--tooltip__trigger bx--tooltip--a11y bx--tooltip--bottom bx--tooltip--align-center"
        showhidebtn.append(
            Icon(
                "view--off",
                _class="bx--icon--visibility-off",
                hidden="true",
                aria_hidden="true",
            )
        )
        showhidebtn.append(
            Icon("view", _class="bx--icon--visibility-on", aria_hidden="true")
        )
        self[1].append(showhidebtn)



import htmlgenerator as hg
from _strptime import TimeRE
from django.utils import formats

from .datetimeformatstring import to_php_formatstr
from .helpers import REQUIRED_LABEL, ErrorList, HelperText, Label
from .icon import Icon


class DatePicker(hg.DIV):
    def __init__(
        self,
        fieldname,
        placeholder="",
        light=False,
        short=False,
        simple=False,
        widgetattributes={},
        boundfield=None,
        **attributes,
    ):
        self.fieldname = fieldname
        attributes["_class"] = attributes.get("_class", "") + " bx--form-item"
        picker_attribs = (
            {}
            if simple
            else {"data-date-picker": True, "data-date-picker-type": "single"}
        )
        widgetattributes["_class"] = (
            widgetattributes.get("_class", "") + " bx--date-picker__input"
        )

        input = hg.INPUT(
            placeholder=placeholder,
            type="text",
            **widgetattributes,
        )
        self.input = input
        if not simple:
            input.attributes["data-date-picker-input"] = True
            input = hg.DIV(
                input,
                Icon(
                    "calendar",
                    size=16,
                    _class="bx--date-picker__icon",
                    data_date_picker_icon="true",
                ),
                _class="bx--date-picker-input__wrapper",
            )

        super().__init__(
            hg.DIV(
                hg.DIV(
                    Label(),
                    input,
                    _class="bx--date-picker-container",
                ),
                _class="bx--date-picker"
                + (" bx--date-picker--simple" if simple else "bx--date-picker--single")
                + (" bx--date-picker--short" if short else "")
                + (" bx--date-picker--light" if light else ""),
                **picker_attribs,
            ),
            **attributes,
        )
        # for easier reference in the render method:
        self.label = self[0][0][0]
        self.simple = simple

        if boundfield is not None:
            if boundfield.field.disabled:
                self.label.attributes["_class"] += " bx--label--disabled"
            self.label.attributes["_for"] = boundfield.id_for_label
            self.label.append(boundfield.label)
            if boundfield.field.required:
                self.label.append(REQUIRED_LABEL)

            dateformat = (
                boundfield.field.widget.format
                or formats.get_format(boundfield.field.widget.format_key)[0]
            )
            dateformat_widget = to_php_formatstr(
                boundfield.field.widget.format,
                boundfield.field.widget.format_key,
            )
            if self.simple:
                self.input.attributes["pattern"] = TimeRE().compile(dateformat).pattern
            else:
                self.input.attributes["data_date_format"] = dateformat_widget

            if boundfield.help_text:
                self[0][0].append(HelperText(boundfield.help_text))
            if boundfield.errors:
                self.input.attributes["data-invalid"] = True
                self[1].append(
                    Icon(
                        "warning--filled",
                        size=16,
                        _class="bx--text-input__invalid-icon",
                    )
                )
                self[0][0].append(ErrorList(boundfield.errors))



from django.contrib.messages.views import SuccessMessageMixin
from django.urls import include, path
from django.utils.translation import gettext_lazy as _
from dynamic_preferences import views as preferences_views
from dynamic_preferences.forms import preference_form_builder
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.users import views as user_preferences_views
from dynamic_preferences.users.registries import user_preferences_registry

from bread.utils import autopath

from .forms.forms import PreferencesForm, UserPreferencesForm
from .views import auth, system, userprofile


# preferences views need a bit of weird treatment, I think we coudl make this code much shorter
def user_get_form_class(self, *args, **kwargs):
    section = self.kwargs.get("section", None)
    return preference_form_builder(
        UserPreferencesForm,
        instance=self.request.user,
        section=section,
        **kwargs,
    )


PreferencesView = type(
    "PreferencesView",
    (SuccessMessageMixin, preferences_views.PreferenceFormView),
    {"success_message": "Preferences updated"},
)


class UserPreferencesView(
    SuccessMessageMixin, user_preferences_views.UserPreferenceFormView
):
    get_form_class = user_get_form_class

    def get_success_message(self, data):
        if data:
            return _("Preferences updated")
        return None


external_urlpatterns = [
    path("ckeditor/", include("ckeditor_uploader.urls")),
    path(
        "preferences/",
        include(
            (
                [
                    path(
                        "global/",
                        PreferencesView.as_view(
                            registry=global_preferences_registry,
                            form_class=PreferencesForm,
                        ),
                        name="global",
                    ),
                    path(
                        "global/<slug:section>",
                        PreferencesView.as_view(
                            registry=global_preferences_registry,
                            form_class=PreferencesForm,
                        ),
                        name="global.section",
                    ),
                    path(
                        "user/",
                        UserPreferencesView.as_view(registry=user_preferences_registry),
                        name="user",
                    ),
                    path(
                        "user/<slug:section>",
                        UserPreferencesView.as_view(registry=user_preferences_registry),
                        name="user.section",
                    ),
                ],
                "dynamic_preferences",
            ),
            namespace="preferences",
        ),
    ),
]


urlpatterns = [
    path("auth/", include("django.contrib.auth.urls")),
    path("accounts/login/", auth.BreadLoginView.as_view(), name="login"),
    path("accounts/logout/", auth.BreadLogoutView.as_view(), name="logout"),
    path("accounts/user/", userprofile.UserProfileView.as_view(), name="userprofile"),
    autopath(
        userprofile.EditPersonalDataView.as_view(), urlname="userprofile.personal"
    ),
    autopath(userprofile.EditLoginView.as_view(), urlname="userprofile.login"),
    autopath(userprofile.password_reset, urlname="userprofile.password_reset"),
    autopath(
        userprofile.EditPermissionsView.as_view(), urlname="userprofile.permissions"
    ),
    autopath(userprofile.set_devmode, urlname="devmode"),
    path(
        "accounts/password_reset/",
        auth.BreadPasswordResetView.as_view(),
        name="password_reset",
    ),
    path(
        "accounts/password_reset/done/",
        auth.BreadPasswordResetDoneView.as_view(),
        name="password_reset_done",
    ),
    path(
        "accounts/reset/<uidb64>/<token>/",
        auth.BreadPasswordResetConfirmView.as_view(),
        name="password_reset_confirm",
    ),
    path(
        "accounts/reset/done/",
        auth.BreadPasswordResetCompleteView.as_view(),
        name="password_reset_complete",
    ),
    path("systeminformation", system.systeminformation, name="systeminformation"),
] + external_urlpatterns



import htmlgenerator as hg

from .helpers import REQUIRED_LABEL, ErrorList, HelperText, Label
from .icon import Icon


class TextArea(hg.DIV):
    def __init__(
        self,
        fieldname,
        placeholder="",
        rows=None,
        cols=None,
        light=False,
        widgetattributes={},
        boundfield=None,
        **attributes,
    ):
        self.fieldname = fieldname
        attributes["_class"] = attributes.get("_class", "") + " bx--form-item"

        widgetattributes["_class"] = (
            widgetattributes.get("_class", "")
            + f" bx--text-area bx--text-area--v2 {'bx--text-area--light' if light else ''}"
        )
        if rows:
            widgetattributes["rows"] = rows
        if cols:
            widgetattributes["cols"] = cols

        super().__init__(
            Label(),
            hg.DIV(
                hg.TEXTAREA(placeholder=placeholder, **widgetattributes),
                _class="bx--text-area__wrapper",
            ),
            **attributes,
        )

        # for easier reference
        self.label = self[0]
        self.input = self[1][0]

        if boundfield.field.disabled:
            self.label.attributes["_class"] += " bx--label--disabled"
            self.input.attributes["disabled"] = True
        if boundfield is not None:
            self.label.attributes["_for"] = boundfield.id_for_label
            self.label.append(boundfield.label)
            if boundfield.field.required:
                self.label.append(REQUIRED_LABEL)

            self.input.append(self.input.attributes.pop("value", ""))

            if boundfield.help_text:
                self.append(HelperText(boundfield.help_text))
            if boundfield.errors:
                self[1].attributes["data-invalid"] = True
                self.input.attributes["_class"] += " bx--text-area--invalid"
                self[1].append(
                    Icon(
                        "warning--filled",
                        size=16,
                        _class="bx--text-area__invalid-icon",
                    )
                )
                self.append(ErrorList(boundfield.errors))



import urllib

from django.contrib.auth import authenticate, login
from django.http import HttpResponseRedirect
from django.urls import reverse


class RequireAuthenticationMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if "bread-cookies" not in request.session:
            request.session["bread-cookies"] = {}
        for k, v in request.COOKIES.items():
            if k.startswith("bread-"):
                request.session["bread-cookies"][k] = urllib.parse.unquote(v)
        response = self.get_response(request)
        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        whitelisted_urlnames = (
            "login",
            "password_reset",
            "password_reset_done",
            "password_reset_confirm",
            "password_reset_complete",
        )
        if request.user.is_authenticated:
            return None
        if request.resolver_match.url_name in whitelisted_urlnames:
            return None
        if "Authorization" in request.headers:
            user = authenticate(request)
            if user is not None:
                login(request, user)
                request.user = user
                return None

        return HttpResponseRedirect(reverse("login") + "?next=" + request.path)



# Generated by Django 3.1.5 on 2021-03-01 14:46

import django.db.models.deletion
from django.db import migrations, models

import bread.contrib.reports.fields.queryfield


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateField(auto_now_add=True, verbose_name='Created')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('filter', bread.contrib.reports.fields.queryfield.QuerysetField(blank=True, modelfieldname='model', verbose_name='Filter')),
                ('model', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.contenttype')),
            ],
            options={
                'verbose_name': 'Report',
                'verbose_name_plural': 'Reports',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ReportColumn',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('column', models.CharField(max_length=255, verbose_name='Column')),
                ('name', models.CharField(max_length=255, verbose_name='Name')),
                ('aggregation', models.CharField(blank=True, choices=[('count', ''), ('sum', '')], max_length=64, verbose_name='Aggregation')),
                ('report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='columns', to='reports.report')),
            ],
        ),
    ]



import htmlgenerator as hg
from django.utils.translation import gettext_lazy as _

from .helpers import REQUIRED_LABEL, ErrorList
from .icon import Icon

# TODO: make delete-field working correctly


class FileUploader(hg.DIV):
    def __init__(
        self,
        fieldname,
        light=False,
        widgetattributes={},
        boundfield=None,
        **attributes,
    ):
        self.fieldname = fieldname
        attributes["_class"] = attributes.get("_class", "") + " bx--form-item"
        widgetattributes["_class"] = (
            widgetattributes.get("_class", "") + " bx--file-input bx--visually-hidden"
        )
        widgetattributes["type"] = "file"
        self.boundfield = boundfield

        self.label = hg.STRONG(_class="bx--file--label")
        self.help_text = hg.P(_class="bx--label-description")
        self.uploadbutton = hg.LABEL(
            hg.SPAN(_("Select file"), role="button"),
            tabindex=0,
            _class="bx--btn bx--btn--primary",
            data_file_drop_container=True,
        )
        self.input = hg.INPUT(
            **widgetattributes,
            onload="""document.addEventListener('change', (e) => {
                this.parentElement.querySelector('[data-file-container]').innerHTML = ''; var widget = new CarbonComponents.FileUploader(this.parentElement); widget._displayFilenames(); widget.setState('edit');
            })""",
        )
        self.container = hg.DIV(data_file_container=True, _class="bx--file-container")
        self.wrapper = hg.DIV(
            self.uploadbutton,
            self.input,
            self.container,
            _class="bx--file",
            data_file=True,
        )
        super().__init__(
            self.label,
            self.help_text,
            self.wrapper,
            **attributes,
        )

    def render(self, context):
        if self.boundfield is not None:
            if self.boundfield.field.disabled:
                self.uploadbutton.attributes["disabled"] = True
                self.input.attributes["disabled"] = True
            self.uploadbutton.attributes["_for"] = self.boundfield.id_for_label
            self.label.append(self.boundfield.label)
            if self.boundfield.field.required:
                self.label.append(REQUIRED_LABEL)
            if self.boundfield.help_text:
                self.help_text.append(self.boundfield.help_text)
            if self.boundfield.errors:
                self.input.attributes["data_invalid"] = True
                self.wrapper.append(ErrorList(self.boundfield.errors))
            if self.boundfield.value():
                self.container.append(
                    hg.SPAN(
                        hg.P(self.boundfield.value(), _class="bx--file-filename"),
                        hg.SPAN(
                            hg.BUTTON(
                                Icon("close", size=16),
                                _class="bx--file-close",
                                type="button",
                                aria_label="close",
                            ),
                            data_for=self.boundfield.id_for_label,
                            _class="bx--file__state-container",
                        ),
                        _class="bx--file__selected-file",
                    )
                )
        return super().render(context)
